
// Local operations on arrays.

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Copy'                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Copy {
public :
  Copy(const DAD* _dst, const DAD* _src) :
           dst(_dst), src(_src), rank(_dst->rnk()) {
    shapeCheck(dst, src) ;

    exts     = new int [rank] ;
    dstSteps = new int [rank] ;
    srcSteps = new int [rank] ;
  }

  ~Copy() {
    delete [] srcSteps ;
    delete [] dstSteps ;
    delete [] exts ;
  }

  void execute(T* dstBase, T* srcBase) {
    if(dst->grp().member())
      loop(dstBase, srcBase, 0) ;
  }

private :
  const DAD* dst ;
  const DAD* src ;
  int rank ;
  int *exts, *dstSteps, *srcSteps ;

  void loop(T* dstBase, T* srcBase, const int r) {
    if(r == rank)
      blockCopy(dstBase, dstSteps, srcBase, srcSteps, exts, rank) ;
    else {
      Stride dstStr = dst->str(r) ;
      Stride srcStr = src->str(r) ;

      LocBlocksIndex i(dst->rng(r)) ;
      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
	exts     [r] = i.count ;
	dstSteps [r] = dstStr.step(i.sub_stp) ;
	srcSteps [r] = srcStr.step(i.sub_stp) ;

	loop(dstBase + dstStr.offset(i), srcBase + srcStr.offset(i), r + 1) ;
      }
    }
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Init'                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Init {
public :
  Init(const DAD* _dst) : dst(_dst), rank(_dst->rnk()) {
    exts  = new int [rank] ;
    steps = new int [rank] ;
  }

  ~Init() {
    delete [] steps ;
    delete [] exts ;
  }

  void execute(T* dstBase, T src) {
    if(dst->grp().member())
      loop(dstBase, 0, src) ;
  }

private :
  const DAD* dst ;
  int rank ;
  int *exts, *steps ;

  void loop(T* dstBase, const int r, T src) {
    if(r == rank)
      blockInit(dstBase, steps, exts, rank, src) ;
    else {
      Stride str = dst->str(r) ;

      LocBlocksIndex i(dst->rng(r)) ;
      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
	exts  [r] = i.count ;
	steps [r] = str.step(i.sub_stp) ;

	loop(dstBase + str.offset(i), r + 1, src) ;
      }
    }
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Spread'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Spread {
public :
  Spread(const DAD* _dst, const DAD* _src, const int _dim) :
           dst(_dst), src(_src),
           dstRank(_dst->rnk()), srcRank(_src->rnk()), dim(_dim) {
    //shapeCheck(dst, src) ;

    exts     = new int [dstRank] ;
    dstSteps = new int [dstRank] ;
    srcSteps = new int [srcRank] ;
  }

  ~Spread() {
    delete [] srcSteps ;
    delete [] dstSteps ;
    delete [] exts ;
  }

  void execute(T* dstBase, T* srcBase) {
    if(dst->grp().member())
      loop(dstBase, srcBase, 0, 0) ;
  }

private :
  const DAD* dst ;
  const DAD* src ;
  int dstRank, srcRank, dim ;
  int *exts, *dstSteps, *srcSteps ;

  void loop(T* dstBase, T* srcBase, const int d, const int r) {
    if(d == dstRank) {
      LocBlocksIndex i(dst->rng(dim)) ;

      Stride dstStr = dst->str(dim) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
	exts     [srcRank] = i.count ;
	dstSteps [srcRank] = dstStr.step(i.sub_stp) ;

	blockVecSpread(dstBase + dstStr.offset(i), dstSteps, exts,
		       srcBase, srcSteps, srcRank) ;
      }
    }
    else {
      if(d == dim)
	loop(dstBase, srcBase, d + 1, r) ;
      else {
	LocBlocksIndex i(dst->rng(d)) ;

	Stride dstStr = dst->str(d) ;
	Stride srcStr = src->str(r) ;

	for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
	  exts     [r] = i.count ;
	  dstSteps [r] = dstStr.step(i.sub_stp) ;
	  srcSteps [r] = srcStr.step(i.sub_stp) ;

	  loop(dstBase + dstStr.offset(i), srcBase + srcStr.offset(i),
	       d + 1, r + 1) ;
	}
      }
    }
  }
} ;


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
////  Class Map: blockwise mapping operation.                            //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class Map {
//public :
//  Map(const DAD* _arr) : arr(_arr), rank(_arr->rnk()) {
//    exts     = new int [rank] ;
//    step = new int [rank] ;
//  }
//
//  ~Map() {
//    delete [] steps ;
//    delete [] exts ;
//  }
//
//  void execute() {
//    if(arr->grp().member())
//      loop(0, 0, 0) ;
//  }
//
//protected :
//  int rank ;
//  int *exts, *steps ;
//
//  virtual void opMap(const int offset) = 0 ;
//
//private :
//  const DAD* arr ;
//
//  void loop(int offset, const int r) {
//    if(r == rank)
//      opMap(offset) ;
//    else {
//      Stride str = arr->str(r) ;
//
//      LocBlocksIndex i(arr->rng(r)) ;
//      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
//        exts  [r] = i.count ;
//        steps [r] = str.step(i.sub_stp) ;
//
//        loop(offset + str.offset(i), r + 1) ;
//      }
//    }
//  }
//} ;

//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
////  Class Map2: blockwise mapping operation applied to pair of         //
////              identically distributed arrays.                        //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class Map2 {
//public :
//  Map2(const DAD* _dst, const DAD* _src) :
//           dst(_dst), src(_src), rank(_src->rnk()) {
//    shapeCheck(dst, src) ;
//
//    exts     = new int [rank] ;
//    dstSteps = new int [rank] ;
//    srcSteps = new int [rank] ;
//  }
//
//  ~Map2() {
//    delete [] srcSteps ;
//    delete [] dstSteps ;
//    delete [] exts ;
//  }
//
//  void execute() {
//    if(src->grp().member())
//      loop(0, 0, 0) ;
//  }
//
//protected :
//  int rank ;
//  int *exts, *srcSteps, *dstSteps ;
//
//  virtual void opMap(const int dstOffset, const int srcOffset) = 0 ;
//
//private :
//  const DAD* dst ;
//  const DAD* src ;
//
//  void loop(int dstOffset, int srcOffset, const int r) {
//    if(r == rank)
//      opMap(dstOffset, srcOffset) ;
//    else {
//      Stride dstStr = dst->str(r) ;
//      Stride srcStr = src->str(r) ;
//
//      LocBlocksIndex i(dst->rng(r)) ;
//      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
//        exts     [r] = i.count ;
//        dstSteps [r] = dstStr.step(i.sub_stp) ;
//        srcSteps [r] = srcStr.step(i.sub_stp) ;
//
//        loop(dstOffset + dstStr.offset(i), srcOffset + srcStr.offset(i),
//             r + 1) ;
//      }
//    }
//  }
//} ;

//template<class T>
//class Copy : public Map2 {
//public :
//  Copy(const DAD* _dst, const DAD* _src) : Map2(_dst, _src) {}
//
//  void execute(T* _dstBase, T* _srcBase) {
//    dstBase = _dstBase ;
//    srcBase = _srcBase ;
//    Map2 :: execute() ;
//  }
//
//  void opMap(const int dstOffset, const int srcOffset) {
//    blockCopy(dstBase + dstOffset, dstSteps, srcBase + srcOffset, srcSteps,
//              exts, rank) ;
//  }
//
//private :
//  T *dstBase, *srcBase ;
//} ;


// -------------------------------------------------------------
// Things to do
//
//   Need a version of `shapeCheck' appropriate to `Spread'.
//
//   Should do contiguity optimizations in `Copy', etc?
//
//   Should `Copy', `Spread' allow destination and source with different
//       (but interconvertable) element types?
